/* lspbs.c 
 *
 * The lspbs utility reports clear, concise and up-to-date node and CPU usage.
 * Copyright (C) 2003-2007 Joseph Spadavecchia <j.spadavecchia@ed.ac.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <pbs_ifl.h>


/* node/cpu status */
#define FREE    0
#define PARTIAL 1
#define USED    2
#define DOWN    3 
#define OFFLINE 4 

struct node {
  unsigned long np;
  unsigned long status;
  unsigned long *cpu_state;
  char **cpu_job;
  char *name;
  struct node *next;
};

struct filter {
  char *value;
  struct filter *and;
  struct filter *or;
};

/* display settings */
#define LINE_LEN  48
#define GREEN     "\033[1;32m"
#define GREY      "\033[1;30m"
#define YELLOW    "\033[1;33m"
#define RED       "\033[1;31m"
#define NORMAL    "\033[1;00m"

#define DEFAULT_MODE    0 
#define LIST_MODE       1

static const char version[] = "lspbs v1.1; Joseph Spadavecchia, October 2006";
static const char usage[]   = "Usage: lspbs [options] \n"
                              "Options:\n"
                              "  -h or --help\t\t\tYou're looking at it.\n"
                              "  -f or --filter <properties>\tFilter nodes by properties.\n"
                              "  -l or --list\t\t\tJust list the nodes.\n"
                              "  -s or --server <pbs-server>\tSelect PBS server.\n"
                              "  -S or --state  <node-state>\tFilter nodes by state.\n"
                              "  -v or --version\t\tShow version information.";

static const char *alive_alias[] = {ND_free, ND_job_exclusive, ND_job_sharing, ND_busy};
static const char *dead_alias[] = {ND_down, ND_offline, ND_state_unknown};


/* Given a string of node attributes @attr and
 * a pointer to a group of @filters determine
 * whether the node should be filtered out.
 *
 * Returns 1 if the node is to be filtered, otherwise 0
 */
int filter_node(char * attr, struct filter *filters);

/* frees the filter group pointed to by @filters */
void filter_cleanup(struct filter *filters);

int main(int argc, char *argv[])
{
  int mode, con, c;
  char *server, *server_opt, *job_id, *state;
  struct batch_status *pbs_node, *pbs_job;
  struct attrl *attr, *job_attr;
  int filter, properties;

  struct node *nodes, newnode, *node;
  struct filter *state_filters, *node_filters, *group, *fltr;
  unsigned long i, j, k;

  unsigned long node_count = 0;
  unsigned long cpu_count = 0;
  unsigned long free_count = 0;
  unsigned long busy_count = 0;
  unsigned long down_count = 0;
  unsigned long offline_count = 0;
  unsigned long max_name_len = 0;
  unsigned long nodes_per_line = 0;
  unsigned long lines = 0;

  mode = DEFAULT_MODE;
  server_opt = NULL;

  nodes = NULL;
  node = NULL;
  state_filters = NULL;
  node_filters = NULL;

  /* parse command line */
  while(1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"help", 0, 0, 'h'},
      {"filter", 1, 0, 'f'},
      {"list", 0, 0, 'l'},
      {"server", 1, 0, 's'},
      {"state", 1, 0, 'S'},
      {"version", 0, 0, 'v'},
      {0, 0, 0, 0}
    };
    
    c = getopt_long(argc, argv, "hf:ls:S:v",
                    long_options, &option_index);
    if(c == -1)
      break;

    switch(c) {
    case 'h':
      fprintf(stderr, "%s\n", version);
      fprintf(stderr, "%s\n\n", usage);
      fprintf(stderr, "For reporting bugs email: <j.spadavecchia@ed.ac.uk>\n");
      exit(EXIT_SUCCESS);
      break;

    case 'f':
      /* check for something */
      optarg = strtok(optarg, ",");
      if(!optarg) break;

      /* add node filters */
      if(!node_filters) {
        node_filters = (struct filter *)malloc(sizeof(struct filter));
        group = node_filters;
      } else {
        for(group = node_filters; group->or; group = group->or);
        group->or = (struct filter *)malloc(sizeof(struct filter));
        group = group->or;
      }
      group->or = NULL;
      fltr = group;

      fltr->value = (char *)malloc(sizeof(char)*strlen(optarg) + 1);
      strcpy(fltr->value, optarg);

      while((optarg = strtok(NULL, ","))) {
        fltr->and = (struct filter *)malloc(sizeof(struct filter));
        fltr = fltr->and;

        fltr->and = NULL;
        fltr->value = (char *)malloc(sizeof(char)*strlen(optarg) + 1);
        strcpy(fltr->value, optarg);
      }
      break;

    case 'l':
      mode = LIST_MODE;
      break;

    case 's':
      server_opt = optarg;
      break;

    case 'S':
      optarg = strtok(optarg, ",");
      /* check for something */
      if(!optarg) break;

      /* add state filters */
      if(!state_filters) {
        state_filters = (struct filter *)malloc(sizeof(struct filter));
        group = state_filters;
      } else {
        for(group = state_filters; group->or; group = group->or);
        group->or = (struct filter *)malloc(sizeof(struct filter));
        group = group->or;
      }
        
      group->or= NULL;
      fltr = group;
        
      fltr->value = (char *)malloc(sizeof(char)*strlen(optarg) + 1);
      strcpy(fltr->value, optarg);

      while((optarg = strtok(NULL, ","))) {
        fltr->and = (struct filter *)malloc(sizeof(struct filter));
        fltr = fltr->and;

        fltr->and = NULL;
        fltr->value = (char *)malloc(sizeof(char)*strlen(optarg) + 1);
        strcpy(fltr->value, optarg);
      }
      break;

    case 'v':
      fprintf(stderr, "%s\n\n", version);
      exit(EXIT_SUCCESS);
      break;

    case '?':
      fprintf(stderr, "See --help for usage information.\n");
      exit(EXIT_FAILURE);
      break;
      
    default:
      fprintf (stderr, "?? getopt returned character code 0%o ??\n", c);
    }
  }
  
  if(optind < argc) {
    fprintf(stderr, "%s: invalid option\n"
            "See --help for usage information.\n", 
            argv[0]);
    exit(EXIT_FAILURE);
  }

  server = server_opt ? server_opt : pbs_default();
  if(!server) {
    fprintf(stderr, "cannot get default server name\n");
    exit(EXIT_FAILURE);
  }

  if(mode == DEFAULT_MODE)
    printf("server: %s%s\n", server, server_opt ? "" : " (default)");

  pbs_node = NULL;
  con = pbs_connect(server);
  if(con > 0) {
    pbs_node = pbs_statnode(con, (char *)NULL, (struct attrl * )NULL, 
                            (char *)NULL);
    if(!pbs_node) {
      fprintf(stderr, "cannot statnode\n");
      exit(EXIT_FAILURE);
    }
  } else {
    fprintf(stderr, "cannot connect to server\n");
    exit(EXIT_FAILURE);
  }
 
  /* collect node and cpu status information */
  for(; pbs_node; pbs_node = pbs_node->next) {
    newnode.next = NULL;
    newnode.cpu_state = NULL;
    newnode.cpu_job = NULL;
    newnode.name = pbs_node->name;
    newnode.status = USED;

    filter = 1;
    properties = 0;
    for(attr = pbs_node->attribs; attr; attr = attr->next) {
      if(strcmp(attr->name, ATTR_NODE_state) == 0) {
        /* apply filters based on state */
        filter = filter_node(attr->value, state_filters);
        if(filter) break;

        /* set the node status, according to lspbs */
        state = strtok(attr->value, ",");
        do {
          if(strcmp(state, ND_free) == 0) {
            newnode.status = FREE;
          } else if(strcmp(state, ND_down) == 0 ||
                strcmp(state, ND_state_unknown) == 0) {
            newnode.status = DOWN;
          } else if(strcmp(state, ND_offline) == 0) {
            newnode.status = OFFLINE;
          }
        } while((state = strtok(NULL, ", ")));
      } else if(strcmp(attr->name, ATTR_NODE_np) == 0) {
        /* per CPU information */
        newnode.np = strtoul(attr->value, NULL, 10);
        newnode.cpu_state = (unsigned long *)malloc(sizeof(unsigned long)*newnode.np);
        memset(newnode.cpu_state, 0, sizeof(unsigned long)*newnode.np);
        newnode.cpu_job = (char **)malloc(sizeof(char *)*newnode.np);
        memset(newnode.cpu_job, 0, sizeof(char *)*newnode.np);
      } else if(strcmp(attr->name, ATTR_NODE_jobs) == 0) {
        /* job attributes for CPU status */
        i = 0;
        job_id = strtok(attr->value, ",");
        do {
          newnode.cpu_state[i] = USED;
          pbs_job = pbs_statjob(con, &(job_id[2]), NULL, NULL);

          job_attr = pbs_job ? pbs_job->attribs : NULL;
          for(; job_attr; job_attr = job_attr->next) {
            if(job_attr->name) {
              if(strcmp(job_attr->name, ATTR_owner) == 0) {
                newnode.cpu_job[i] = job_attr->value;
              }
            }
          }
          i++;
        } while((job_id = strtok(NULL, ", ")));

        /* modify the node status based on CPU usage */
        if(newnode.status != DOWN &&
           newnode.status != OFFLINE) {
          for(i = 0; i < newnode.np; i++) {
            if(newnode.cpu_state[i] != USED) {
              newnode.status = PARTIAL;
            }
          }
        }
      } else if(strcmp(attr->name, ATTR_NODE_properties) == 0) {
        properties = 1;

        /* filters based on node properties */
        filter = filter_node(attr->value, node_filters);

        if(filter) break;
      }
    }

    if(node_filters && !properties) filter = 1;

    if(!filter) {
      if(strlen(pbs_node->name) > max_name_len)
        max_name_len = strlen(pbs_node->name);

      if(!node_count++) {
        node = (struct node *)malloc(sizeof(struct node));
        *node = newnode;
        nodes = node;
      } else {
        node->next = (struct node *)malloc(sizeof(struct node));
        node = node->next;
        *node = newnode;
      }
    } else {
      if(newnode.cpu_state) free(newnode.cpu_state);
      if(newnode.cpu_job) free(newnode.cpu_job);
    }

  }

  /* for list mode only */
  if(mode == LIST_MODE) {
    for(node = nodes; node; node = node->next) {
      printf("%s ", node->name);
      if(!node->next) printf("\n");
    } 
    goto cleanup;
  }

  /* print out pretty usage table */
  lines = ceil((double)node_count*(max_name_len + 2)/(double)LINE_LEN) + 1;
  nodes_per_line = LINE_LEN/(max_name_len + 1);
  
  printf("key: free " GREEN "*" NORMAL ", shared "  YELLOW "*" NORMAL 
         ", busy *, offline " GREY "*" NORMAL ", down " RED "*" NORMAL "\n\n");
  
  node = nodes;
  for(i = 0; node && i < lines; i++) {
    int len;
    for(j = 0; node && j < nodes_per_line ; j++) {
      switch(node->status) {
      case FREE: 
        printf(GREEN);
        break; 

      case PARTIAL:
        printf(YELLOW);
        break;

      case USED:
        printf(NORMAL);
        break;

      case DOWN: 
        printf(RED);
        break;

      case OFFLINE: 
        printf(GREY);
        break;
      }
      printf("%s" NORMAL, node->name);

      len = strlen(node->name);
      for(k = 0; k < max_name_len + 1 - len; k++)
        printf(" ");

      node = node->next;
    }
    printf("\n");
  }

  /* gather some stats and print footer */
  for(node = nodes; node; node = node->next) {
    cpu_count += node->np;
    switch(node->status) {
    case FREE:
      free_count += node->np;
      break;

    case DOWN:  
      down_count += node->np;
      break;
      
    case OFFLINE: 
      offline_count += node->np;
      break;

    case PARTIAL:
      for(j = 0; j < node->np; j++) {
        if(node->cpu_state[j] == USED)
          busy_count++;
        else 
          free_count++;
      }
      break;

    case USED:
      busy_count += node->np;
      break;
    }
  }

  printf("\n");
  printf("%lu CPUs: %lu free, %lu busy, %lu offline, %lu down\n\n", 
         cpu_count, free_count, busy_count, offline_count, down_count);
 
cleanup:

  /* cleanup node info */
  for(node = nodes; node;) {
    free(node->cpu_state);
    free(node->cpu_job);
    nodes = node->next;
    free(node);
    node = nodes;
  }

  /* cleanup filters */
  filter_cleanup(node_filters);
  filter_cleanup(state_filters);

  pbs_disconnect(con);

  return EXIT_SUCCESS;
}

int filter_node(char * attr, struct filter *filters)
{
  int filter = 0;
  size_t i;
  struct filter *group, *fltr;
  char *p;

  /* if any of the filters in the group pass then
     the node is not filtered */
  for(group = filters; group; group = group->or) {
    filter = 0;
    /* if any of the members of this filter fails
       then set the filter flag, but try the next */
    for(fltr = group; fltr; fltr = fltr->and) {

      /* apply aliases */
      if(strcmp(fltr->value, "alive") == 0) {
        for(i = 0; i < sizeof(alive_alias)/sizeof(char *) && !(p = strstr(attr, alive_alias[i])); i++);
      } else if(strcmp(fltr->value, "dead") == 0) {
        for(i = 0; i < sizeof(dead_alias)/sizeof(char *) && !(p = strstr(attr, dead_alias[i])); i++);
      } else {
        p = strstr(attr, fltr->value);
      }

      /* check filter */
      if(!p ||
         ((&(p[0]) != &(attr[0]) && p[strlen(fltr->value)] != '\0') &&
         (p[0] != ',' && p[strlen(fltr->value)] != ',' ) &&
         (p[0] != ',' && p[strlen(fltr->value)] != '\0' ))
        ) {
        filter = 1;
        break;
      }
    }
    if(!filter) break;
  }

  return filter;
}

void filter_cleanup(struct filter *filters)
{
  struct filter *group, *fltr, *tmp;
  for(group = filters; group;) {
    for(fltr = group->and; fltr;) {
      tmp = fltr->and;
      free(fltr);
      fltr = tmp;
    }

    filters = group->or;
    free(group);
    group = filters;
  }
}
